#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "history.h"

/******************************************************
 *   history * history_create(int id);
 */

history * history_create(int id){
	history * h;
	h = malloc(sizeof(history));
	assert(h);
	h->id=id;
	h->trust=EPSILON;
	h->nGood=h->nInteractions=0;
	h->blocksStored=0;
	h->shouldQuery=1;
	h->badguy=0;
	h->savedGetRequest=NULL;
	return h;
}

/******************************************************
 */
void history_free(history* h){
	if(h->savedGetRequest)
		free(h->savedGetRequest);
	free(h);
}

/******************************************************
 * int history_update(context * ctx, int id, int dst_id,int result);
 *	find a node's targets, and update it with a new
 *	interaction.  If there is no existing history, and there
 *	is room, create one.
 */


int history_update(context * ctx, int id, int dst_id,int result){
	node *src;
	int i;
	
	src = ctx->nodes[id];
	for(i=0;i<src->nTargets;i++){
		if(src->targets[i]->id == dst_id)
			break;
	}
	if(i>=src->nTargets){	// didn't find one
		if(src->nTargets<src->maxTargets){	// try to add
			i=src->nTargets;
			src->targets[src->nTargets++]=
				history_create(dst_id);

		} else return -1;	// failed to add
	}
	src->targets[i]->nInteractions++;
	src->targets[i]->nGood+=result;
	src->targets[i]->trust = (double)src->targets[i]->nGood/
		src->targets[i]->nInteractions;
	src->targets[i]->shouldQuery=1;	// stuff has changed with them
	if((src->targets[i]->nInteractions-src->targets[i]->nGood)>ctx->maxbad)
		src->targets[i]->badguy=1;	// brand the evil doer!!	// only useful for strawman
	return 0;
}

/***********************************************************
 * int history_good_cookie_count(context *ctx, int id);
 * 	return the number of gotGoodCookie fields set
DEPRECATED
int history_good_cookie_count(context *ctx, int id){
	int i;
	int count=0;
	for(i=0;i<ctx->nodes[id]->nTargets;i++){
		if(ctx->nodes[id]->targets[i]->gotGoodCookie)
			count++;
	}
	return count;
}

 */

/******************************************************
 * int history_add_with_trust(context * ctx, int id, int dst_id,trust)
 *	if a history exists, no nothing, else create one if there is space
 *	fill in the trust value
 */


int history_add_with_trust(context * ctx, int id, int dst_id, double trust){
	node *src;
	int i;
	
	src = ctx->nodes[id];
	if(id==dst_id)	// never add yourself to your own history
		return 0;
	for(i=0;i<src->nTargets;i++){
		if(src->targets[i]->id == dst_id)
			break;
	}
	// if we didn't find it, and there is room, add it
	if((i>=src->nTargets)&& (src->nTargets<src->maxTargets)){
		i=src->nTargets;
		src->targets[src->nTargets]= history_create(dst_id);
		src->targets[src->nTargets++]->trust=trust;
	} else return -1;	// failed to add
	return 0;
}

/******************************************************
 * int history_remove(struct context * ctx, int src_id, int dst_id);
 *	remove a history of someone, presumably b/c they are bad
 */

int history_remove(struct context * ctx, int src_id, int dst_id){
	int i;
	node *n;
	n=ctx->nodes[src_id];
	for(i=0;i<n->nTargets;i++){
		if(n->targets[i]->id==dst_id)
			break;
	}
	if(i>=n->nTargets)
		return -1;		// didn't find them - but that's okay
	n->nTargets--;
	history_free(n->targets[i]);
	n->targets[i]=n->targets[n->nTargets];	// replace with the last in the list
	return 0;
}

/**********************************************************
 * int history_flush_and_refill(struct context *, int id);
 *	get rid of any history entries that aren't useful
 *	get some more people from topo server
 */

int history_flush_and_refill(context *ctx, int id){
	node *n;
	int i;
	int count;
	int buf[BUFLEN];
	int init;

	n = ctx->nodes[id];
	for(i=0;i<n->nTargets;i++){
		if(ctx->nodes[n->targets[i]->id]->done)
			history_remove(ctx,id,n->targets[i]->id);
	}
	if(n->nTargets>=n->maxTargets)
		return 0;	// don't need to be flushed
	init= n->nTargets;
	count=MIN(BUFLEN,n->maxTargets - n->nTargets);	// replace ones we removed
	if(count==0)
		return 1;
	toposerver_get(ctx->ts,id,buf,&count);
	toposerver_add(ctx->ts,id);
	for(i=0;i<count;i++){
		if(trust_calc_back_cookie_path(ctx,id,buf[i])>0)
			history_add_with_trust(ctx,id,buf[i],EPSILON);
	}
	fprintf(ctx->logfile,"%f - history_flush_and_refill:: id %d  - start %d diff %d\n",
			ctx->now, id, init, count);
	return count;
}



/*******************************************************
 * historycmp:
 * 	return x>0 if a>b
 * 		x<0 if a<b
 * 		x=0 otherwise
 */

int historycmp(const void *a, const void *b){
	history *h1, *h2;
	h1=*(history **)a;
	h2=*(history **)b;
	// can't do a slick subtract b/c they are floats and could get
	// truncated to zero or some such
	if(h1->trust>h2->trust)
		return -1;
	if(h2->trust>h1->trust)
		return 1;
	return (h2->nInteractions-h1->nInteractions);	// tie breaker
}


/********************************************************
 * histories_sort_by_trust:
 * 	just do it
 */

int histories_sort_by_trust(context * ctx, node *n){
	int i,j;
	history *h;
	for(i=0;i<n->nTargets;i++){		// mix things up so ties are random
		j = random()%n->nTargets;
		h = n->targets[j];
		n->targets[j]=n->targets[i];
		n->targets[i]=h;
	}
	qsort(n->targets,n->nTargets,sizeof(history *),historycmp);
	return 0;
}

/*******************************************************
 * int history_get_max_transactions(struct context *, history *);
 * 	based on the current rules from the context, return the
 * 	max number of transactions we allow with this node
 */

int history_get_max_transactions(context * ctx, history *h){

	switch(ctx->maxTransactionsRule){
		case 0:
			if(h==NULL)		// if we don't know them
				return 1;	// just let them try once
			return h->nGood+1;	// dumb rule
		case 1:
			if(h==NULL)
				return 1;	// just 1 for now to get them hooked
			if((h->nGood>0)&&(h->nGood>=h->nInteractions))
				return (int)pow(2,h->nGood+1);	// make 100% good people increase exponentially
			else
				return h->nGood+1;
		default:
			fprintf(stderr,"history_get_max_transactions: Bad max rule %d\n",ctx->maxTransactionsRule);
			abort();
	};
	return -1;
}


/******************************************************
 * int history_set_abs_resp(struct context * ctx, int src_id, int dst_id, int resp);
 * 	remember how much they said we stored in them
 */

int history_set_abs_resp(struct context * ctx, int src_id, int dst_id, int resp){
	node *n;
	int i;
	n=ctx->nodes[src_id];

	for(i=0;i<n->nLocations;i++){
		if(n->locations[i]->id == dst_id)
			break;
	}
	if(i>=n->nLocations){	// if we didn't find them, get rid of someone and make a new history 
		if(i>=n->maxLocations){
			for(i=0;i<n->nLocations;i++){
				if(n->locations[i]->blocksStored==0)
					break;
			}
			if(i>=n->maxLocations)
				return -1;	// can't find anyone to oust
			history_free(n->locations[i]);	// out with old
		} else 
			n->nLocations++;
		
		n->locations[i]=history_create(dst_id);	// in with new

	}
	n->locations[i]->blocksStored=resp;
	return 0;
}


/******************************************************
 * int history_set_resp(struct context * ctx, int src_id, int dst_id, int resp);
 * 	add to how much they said we stored in them
 */


int history_set_resp(struct context * ctx, int src_id, int dst_id, int resp){
	node *n;
	int i;
	n=ctx->nodes[src_id];

	for(i=0;i<n->nLocations;i++){
		if(n->locations[i]->id == dst_id)
			break;
	}
	if(i>=n->nLocations){	// if we didn't find them, get rid of someone and make a new history 
		if(i>=n->maxLocations){
			for(i=0;i<n->nLocations;i++){
				if(n->locations[i]->blocksStored==0)
					break;
			}
			if(i>=n->maxLocations)
				return -1;	// can't find anyone to oust
			history_free(n->locations[i]);	// out with old
		} else 
			n->nLocations++;
		n->locations[i]=history_create(dst_id);	// in with new

	}
	n->locations[i]->blocksStored+=resp;
	return 0;
}


/******************************************************
 * int history_get_resp(struct context * ctx, int src_id, int dst_id, int resp);
 * 	return how much they said we stored in them
 */

int history_get_resp(struct context * ctx, int src_id, int dst_id){
	node *n;
	int i;

	n=ctx->nodes[src_id];
	for(i=0;i<n->nLocations;i++){
		if(n->locations[i]->id == dst_id)
			break;
	}
	if(i<n->nLocations)		// make sure we found someone
		return n->locations[i]->blocksStored;
	else
		return 0;
}


/******************************************************
 * history * history_lookup(context *ctx, int id, int dst_id)
 *	look up someone's history
 */

history * history_lookup(context *ctx, int id, int dst_id){
	int i;
	node *n = ctx->nodes[id];
	for(i=0;i<n->nTargets;i++){
		if(n->targets[i]->id==dst_id)
			return n->targets[i];
	}
	return NULL;
}
