#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <signal.h>


#include "pqueue.h"
#include "context.h"
#include "node.h"

#define BIGBUFSIZE 65536*10

void parse_args(context *,int, char * []);
void read_config(context *);
void init(context *);
void usage(context *,char * , char*);
void stubconfig(context *);
void newround(context * ctx);
void printstats(context * ctx);
void raise_sigtrap();
void heartbeat(context *ctx, event *e, double t);
void cleanup(context *ctx);



/*********************************************************************/

int main(int argc, char * argv[]){
	double t;
	event * e;
	context * ctx;
	int i;
	
	ctx = context_create();
	parse_args(ctx, argc,argv);
	read_config(ctx);
	init(ctx);
	// run simulator
	for(i=0;i<ctx->nRounds;i++){
		ctx->curRound=i;
		while((e=(event *)pqueue_extract(ctx->pq,&t))!=NULL){
			ctx->now=t;
			if(ctx->stopAt==t)
				raise_sigtrap();
			e->fun(ctx, e,t);
			if((ctx->totGood<=ctx->doneCount)||(ctx->endAt< t))
				cleanup(ctx);
		}
		heartbeat(ctx,NULL,ctx->now);
		printstats(ctx);
		//newround(ctx);
	}
	free(ctx->ts->nodeIds);
	free(ctx->ts);
	free(ctx->pq->array);
	free(ctx->pq);
	free(ctx);
	return 0;
}

/*********************************************************************/

void parse_args(context * ctx,int argc, char * argv[]){
	int i;
	for(i=1;i<argc;i++){
		if(!strcmp(argv[i],"-s")){		// set the seed
			i++;
			if(i>=argc) usage(ctx," -s needs an arg",NULL);
			ctx->seed=atol(argv[i]);
			continue;
		} else if(!strcmp("-config",argv[i])){
			i++;
			if(i>=argc) usage(ctx," -config needs an arg",NULL);
			strncpy(ctx->configFileName,argv[i],BUFLEN);
			continue;
		} else if(!strcmp("-l",argv[i])){
			ctx->useLogFile=1;
			continue;
		} else if(!strcmp("-L",argv[i])){
			ctx->useLogFile=2;
			continue;
		} else usage(ctx,"Unknown arg", argv[i]);
	}
	return;
}
/*********************************************************************/
void usage(context * ctx, char * s1, char * s2){
	if(s1)
		fprintf(stderr,"%s",s1);
	if(s2)
		fprintf(stderr," %s",s2);
	if(s1||s2)
		fprintf(stderr,"\n");
	fprintf(stderr,"hiertrust [options]\n"
			"	-l :	log data[stdout]\n"
			"	-L :	log NO data[off]\n"
			"	-s seed :	randomseed[%ld]\n"
			"	-config f :	configfile[none]\n"
			"\n",
			ctx->seed);
	exit(1);
}
/*********************************************************************/

void read_config(context * ctx){
	int i;
	FILE * config;
	char buf[BUFLEN];
	char *bigbuf;
	double * dptr;
	double good;
	int want,have,state,n,sleep,tmpi;
	node ** tmp;

	if(!strcmp("",ctx->configFileName))
		return stubconfig(ctx);
	// open the output logging file as a function of the config name
	if(ctx->useLogFile==1)
		snprintf(buf,BUFLEN,"%s.out",ctx->configFileName);
	else 
		snprintf(buf,BUFLEN,"/dev/null");
	if(ctx->useLogFile>0){
		bigbuf= malloc(BIGBUFSIZE);
		assert(bigbuf);
		ctx->logfile=fopen(buf,"w");
		setvbuf(ctx->logfile,bigbuf,_IOFBF,BIGBUFSIZE);
		assert(ctx->logfile);
	}
	if(ctx->useLogFile){
		snprintf(buf,BUFLEN,"%s.stat",ctx->configFileName);
		bigbuf= malloc(BIGBUFSIZE);
	        assert(bigbuf);
		ctx->statfile=fopen(buf,"w");
                setvbuf(ctx->statfile,bigbuf,_IOFBF,BIGBUFSIZE);
                assert(ctx->statfile);
	}

	// parse the config file
	config = fopen(ctx->configFileName,"r");
	if(!config){
		perror("fopen config");
		exit(1);
	}
	while(fgets(buf,BUFLEN,config)){
		if(strncmp(buf,"#",1)==0)	// skip comments
			continue;
		else if(sscanf(buf,"NODES %d good=%lf want=%d have=%d state=%d sleep=%d",
					&n,&good,&want,&have,&state,&sleep)){
			if(ctx->nNodes==0){	// first NODES statement
				ctx->nNodes=0;
				ctx->nodes=malloc(n*sizeof(node *));
				assert(ctx->nodes);
			}else {	// subsequent NODES statement
				tmp=ctx->nodes;	
				ctx->nodes = malloc((n+ctx->nNodes)*sizeof(node *));
				memcpy(ctx->nodes,tmp,sizeof(node *)*ctx->nNodes);
			}
			for(i=ctx->nNodes;i<(ctx->nNodes+n);i++){
				ctx->nodes[i]=node_create(ctx,want,have,state,good,sleep);
				assert(ctx->nodes[i]);
				//event_create(ctx,node_join,-1,ctx->nodes[i]->id,NULL,(double)i/(ctx->nNodes+n));
				event_create(ctx,node_join,-1,ctx->nodes[i]->id,NULL,(double)random()/RAND_MAX);
			}
			ctx->nNodes+=n;
			if(good>GOODTHRESH)
				ctx->totGood+=n;
		} else if(sscanf(buf,"HEARTBEAT %lf",&good)){
			fprintf(stderr,"HEARTBEAT every %lf\n",good);
			dptr=malloc(sizeof(double));
			*dptr=good;
			event_create(ctx,heartbeat,-1,-1,dptr,good);
		} else if(sscanf(buf,"ENDAT %lf",&good)){
			fprintf(stderr,"Going to end at %lf\n",good);
			ctx->endAt=good;
		} else if(sscanf(buf,"STOPAT %lf",&good)){
			fprintf(stderr,"Going to stop at %lf\n",good);
			ctx->stopAt=good;
		} else if(sscanf(buf,"MAXQUERIES %d",&ctx->maxQueries)){
			fprintf(stderr,"Running %d maxQueries \n",ctx->maxQueries);
		} else if(sscanf(buf,"ROUNDS %d",&ctx->nRounds)){
			fprintf(stderr,"Running %d rounds of the sim\n",ctx->nRounds);
		} else if(sscanf(buf,"COOKIES %d ",&tmpi)){
			ctx->cookieProtocol=tmpi;
			if(ctx->cookieProtocol)
				fprintf(stderr,"Using the cookies protocol\n");
		} else if(sscanf(buf,"MAXRULE %d ",&tmpi)){
			ctx->maxTransactionsRule=tmpi;
			fprintf(stderr,"Using max transaction rule %d\n",ctx->maxTransactionsRule);
		} else if(sscanf(buf,"BACKOFF %lf ",&ctx->backoffProb)){
			fprintf(stderr,"Using backoff probability %f\n",ctx->backoffProb);
		} else {
			fprintf(stderr,"Unknown command '%s'\n",buf);
			exit(1);
		}
	}


}


/*********************************************************************/
void stubconfig(context * ctx){
	int i;
	// stub read_config()
	ctx->nNodes=10;
	ctx->nodes=malloc(ctx->nNodes*sizeof(node *));
	assert(ctx->nodes);
	for(i=0;i<ctx->nNodes;i++){
		ctx->nodes[i]=node_create(ctx,9,10,ctx->statePerNode,1,0);
		assert(ctx->nodes[i]);
		event_create(ctx,node_join,-1,ctx->nodes[i]->id,NULL,(double)i/ctx->nNodes);
	}
}

/*********************************************************************/
void init(context * ctx){
	// seed random number
	srandom(ctx->seed);
	ctx->ts=toposerver_create(ctx->tsState);
}
/*********************************************************************/

void printstats(context *ctx){
	char buf[BUFLEN];

	snprintf(buf,BUFLEN,"STATS: round %d : %d nodes failed %d jobsFailed %d nodes satisfied %d jobs dropped %d tot  %f\n",
			ctx->curRound, ctx->nFailed,ctx->nFailedJobs,ctx->nSatisfied,ctx->jobsDropped, ctx->totJobs, ctx->now);
	fprintf(ctx->logfile,"%s",buf);
	fprintf(stderr,"%s",buf);
}
/*********************************************************************/

void newround(context * ctx){
	int i,j;
	node *n;
	ctx->nFailed=ctx->nFailedJobs=ctx->nSatisfied=ctx->jobsDropped=0;
	abort();	// FIXME need to clear all stats, not done yet
	for(i=0;i<ctx->nNodes;i++){
		n=ctx->nodes[i];
		n->nStores=0;
		for(j=0;j<n->nTargets;j++){
			n->targets[j]->blocksStored=0;
			n->targets[j]->shouldQuery=1;
		}
		n->nMyBlocks=n->totMyBlocks;
		n->nFreeSpace=n->totFreeSpace;
		n->outstandingGets=0;
		n->failAt=ctx->failTime;	
		// resched the nodes randomly
		ctx->now=0;
		event_create(ctx,node_schedule,-1,n->id,NULL,(double)random()/RAND_MAX);
	}
}

/************************************************************************
 * raise_sigtrap(context *ctx, event e, double t)
 * 	durh...
 */

void raise_sigtrap(){
	raise(SIGTRAP);
}


/***********************************************************************
 *  void heartbeat(context *ctx, event *e, double t);
 *		print stats every heartbeat, and then reschedule
 */
 
void heartbeat(context *ctx, event *e, double t){
	double *beattime;
	static int oldDropped=0;
	static int oldTotal=0;

	assert(ctx);
	fprintf(ctx->statfile,"%f - HEARTBEAT %d %d %lf %d %d %lf - nodes failed %d jobsFailed %d nodes satisfied %d \n",
		t, ctx->jobsDropped, ctx->totJobs, 
		ctx->totJobs?(double)ctx->jobsDropped/ctx->totJobs:0,
		ctx->jobsDropped-oldDropped,
		ctx->totJobs-oldTotal,
		(ctx->totJobs-oldTotal)?(double)(ctx->jobsDropped-oldDropped)/(ctx->totJobs-oldTotal):0,
		ctx->nFailed, ctx->nFailedJobs, ctx->nSatisfied);
	fflush(ctx->statfile);
	oldDropped=ctx->jobsDropped;
	oldTotal=ctx->totJobs;
	if(e!=NULL){	// we call this the last time with e==NULL
		beattime = (double *) e->arg;
		if(!pqueue_isempty(ctx->pq))
			event_create(ctx,heartbeat,-1,-1,beattime,ctx->now+*beattime);
		e->arg=NULL;
		event_free(e);
	}
}
/*******************************************************************
 * void cleanup(context * ctx)
 * 	make everything else fail
 */

void cleanup(context * ctx){
	int i;
	for(i=0;i<ctx->nNodes;i++){
		if(ctx->nodes[i]->done)
			continue;
		node_print_status(ctx,ctx->nodes[i]->id);
		if(ctx->nodes[i]->good<GOODTHRESH)	// nothing to do
			continue;			// for bad nodes
		ctx->nFailed++;
		ctx->nFailedJobs+=ctx->nodes[i]->nMyBlocks;
	}
	ctx->pq->size=0;	// cut the queue off
}
