#include <stdio.h>
#include <assert.h>


#include "node.h"
#include "event.h"
#include "query.h"
#include "link.h"


/*******************************************************
 * node_create: alloc a node, and fill in all of the values
 */

node * node_create(context * ctx, int myblocks, int freespace, int maxstate, double good, int sleeperCount){
	static int nodeIdCounter=0;
	node * n;
	n = malloc(sizeof(node));
	assert(n);
	n->id=nodeIdCounter++;
	n->isGroup=0;
	n->nMyBlocks=n->totMyBlocks=myblocks;
	n->nFreeSpace=n->totFreeSpace=freespace;
	n->nLocations=0;
	n->maxLocations=n->nMyBlocks;
	n->locations = malloc(sizeof(history *)*n->maxLocations);
	n->nTargets=0;
	n->maxTargets=ctx->maxQueries;
	n->targets = malloc(sizeof(history *)*n->maxTargets);
	n->outstandingGets=0;
	n->nStores=0;
	n->good=good;
	n->sleeperCount=sleeperCount;
	n->goodcookies = truststore_create();
	n->badcookies = truststore_create();
	n->maxCookies=maxstate;
	n->nTransitQueries=0;
	n->qid=0;
	n->transitQueries=malloc(sizeof(cookieQueryInfo)*MAXTRANSITQUERIES);
	n->done=0;
	n->failAt=ctx->failTime;		// some large amount time 
	n->jobsDropped=0;
	n->totJobs=0;
	assert(n->transitQueries);
	assert(n->locations);
	assert(n->targets);
	return n;
}

/******************************************************************************
 * void node_print_status(context *ctx, int id);
 * 	print the final status of the node
 */

void node_print_status(context *ctx, int id){
	node *n = ctx->nodes[id];

	if(n->nMyBlocks>0){
		fprintf(ctx->logfile,"%.30f : STATUS:: node %d FAILED to "
				"place all blocks: %d left, %d targets: cookies %d %d d/t %d %d SCHED\n",
				ctx->now,n->id,n->nMyBlocks,n->nTargets,
				n->goodcookies->nodes, n->badcookies->nodes,
				n->jobsDropped,n->totJobs);
	} else {
		fprintf(ctx->logfile,"%.30f : STATUS:: node "
				"%d is SATISFIED at %.30f cookies %d %d d/t %d %d \n",
				ctx->now,id,ctx->now, n->goodcookies->nodes, n->badcookies->nodes,
				n->jobsDropped,n->totJobs);
	}
	ctx->doneCount++;
	fflush(ctx->logfile);
}

/********************************************************
 * node_free: free all resources w/ a node
 */

void node_free(node * n){
	int i;
	for(i=0;i<n->nLocations;i++)
		history_free(n->locations[i]);
	free(n->locations);
	for(i=0;i<n->nTargets;i++)
		history_free(n->targets[i]);
	free(n->targets);
	free(n);
}

/********************************************************
 * void node_store_blocks(struct context *, int id, int src_id, int n);
 * 	record that we are storing n blocks for src_id
 */

int node_store_blocks(context * ctx, int id, int src_id, int count){
	node * n;
	int i;
	n=ctx->nodes[id];
	assert(count>0);
	for(i=0;i<n->nStores;i++)
		if(n->stores[i].id == src_id)
			break;
	if(i<n->nStores){
		n->stores[i].nBlocks+=count;
		return 0;
	}
	assert(n->nStores<MAXBLOCKS);
	n->stores[n->nStores].id=src_id;
	n->stores[n->nStores++].nBlocks=count;
	return 0;
}



/********************************************************
 * void node_count_stored_blocks(struct context *, int id, int src_id);
 * 	count  the blocks we are storing for src_id
 */

int node_count_stored_blocks(context * ctx, int id, int src_id){
	node * n;
	int i;
	n=ctx->nodes[id];
	for(i=0;i<n->nStores;i++)
		if(n->stores[i].id == src_id)
			break;
	if(i<n->nStores)
		return n->stores[i].nBlocks;
	return 0;	// aren't storing anything for them
}


/********************************************************
 * void node_drop_blocks(struct context *, int id, int src_id);
 * 	disgard what we were storing for src_id
 */

int node_drop_blocks(context * ctx, int id, int src_id){
	node * n;
	int i;
	n=ctx->nodes[id];
	for(i=0;i<n->nStores;i++)
		if(n->stores[i].id == src_id)
			break;
	if(i>=n->nStores)
		return -1;	// nothing to drop
	n->stores[i].id=0;	// doesn't matter, but helps with debugging
	n->stores[i].nBlocks=0;
	n->nStores--;
	if(n->nStores>0)	// overwrite with the last element
		n->stores[i]=n->stores[n->nStores];
	return 0;	// aren't storing anything for them
}


/***********************************************************
 * node_join: initial function to call upon joining
 */

void node_join(context * ctx, event * e, double t){
	int * list=NULL;
	int len;
	node *src;
	int i;

	list=malloc(sizeof(int)*ctx->nodes[e->id]->maxTargets);
	assert(list);

	src=ctx->nodes[e->id];
	len = src->maxTargets;
	toposerver_get(ctx->ts,e->id,list,&len);
	toposerver_add(ctx->ts,e->id);
	fprintf(ctx->logfile,"-- node %d joined at time %.30f - %d neighbors\n",e->id, t, len);
	for(i=0;i<len;i++)
		src->targets[i]=history_create(list[i]);
	src->nTargets=len;
	event_free(e);
	event_create(ctx,node_schedule,-1,src->id,NULL,ctx->now+EPSILON);
}

/*************************************************************
 * node_schedule:
 * 	fireoff a list of set of space queries so we can figure out 
 * 	who has space for our stuff
 *
 * 	- query the people in order of how much we trust them
 *	- stop when we think we've queried enough to finish our work load
 *	- skip anyone we have already queried before, unless stuff has changed
 */

void node_schedule(struct context * ctx, struct event * e, double t){
	node *src;
	int i,j,max;
	getquery * gq;

	src=ctx->nodes[e->id];
	if(src->done)		// we've already marked this node as done;
		return;		// just ignore any future scheduling requests as it screws up the stats
	fprintf(ctx->logfile,"%.30f : node_schedule:: node %d : %d blocks of %d to alloc %d targets\n",
			e->t,e->id, src->nMyBlocks, src->nFreeSpace, src->nTargets);
	assert(src->nMyBlocks<=src->nFreeSpace);
	// test to see if we have failed
	if(ctx->now>src->failAt){
		node_print_status(ctx,src->id);
		src->done=1;
		if(src->good>GOODTHRESH){
			ctx->nFailed++;
			ctx->nFailedJobs+=src->nMyBlocks;
		}
		return;
	} 
	// test to see if we have succeeded
	if(src->nMyBlocks<=0){
		node_print_status(ctx,src->id);
		src->done=1;
		if(src->good>GOODTHRESH)
			ctx->nSatisfied++;
		return;
	}
	// else try to schedule some jobs
	src->nRequestedBlocks=0;
	histories_sort_by_trust(ctx,src);
	assert(src->outstandingGets==0);
	for(i=0;i<src->nTargets;i++){
		assert(!src->targets[i]->badguy);
		if(src->nRequestedBlocks>=src->nMyBlocks)
			break;	// we've allocated all of our blocks
		max = history_get_max_transactions(ctx,src->targets[i]);
		if(!src->targets[i]->shouldQuery)
			continue;
		// count stuff we already stored with them
		max-= history_get_resp(ctx,e->id,src->targets[i]->id);
		max=MAX(0,max);		// don't allow negative values
		max = MIN(max,src->nMyBlocks-src->nRequestedBlocks);
		max = MIN(max, src->nFreeSpace-src->nRequestedBlocks);
		if(max<=0)
			continue;		// wait for pending check query to return
		assert(max>0);		// we need something to offer
		gq = getquery_create(max,max);
		// try to find a path to them, if its local or precomputed, go on, else come back later
		if(trust_calc_forw_cookie_path(ctx,e->id,src->targets[i]->id,gq)==1)
			continue;
		if(trust_calc_back_cookie_path(ctx,e->id,src->targets[i]->id)<0){
			history_remove(ctx,e->id,src->targets[i]->id);
			continue;
		}
			
		src->targets[i]->shouldQuery=0;
		assert(gq);
		getquery_advertise(gq,e->id);	// always advertise yourself
		for(j=0;j<src->nTargets;j++){
			if(src->targets[j]->trust>EPSILON)	// only advertise good nodes
				getquery_advertise(gq,src->targets[j]->id);
		}
		// figure out the cookie path from us to them
		// queue up the event
		event_create(ctx,node_handleget,src->id,
				src->targets[i]->id,gq,ctx->now+linkdelay(ctx));
		src->nRequestedBlocks+=max;
		src->outstandingGets++;
	}
	if(src->nRequestedBlocks<=0){	// failed to schedule jobs, try again later
		fprintf(ctx->logfile,"%.30f : node_schedule:: not enough "
				"nodes to statisfy node %d : %d out of %d short\n",
				e->t,src->id, src->nMyBlocks-src->nRequestedBlocks,src->nMyBlocks);
		history_flush_and_refill(ctx,src->id);
		event_create(ctx,node_schedule,-1,e->id, NULL,ctx->now+scheduledelay(ctx));
		return;
	} 

	event_free(e);
}

/*********************************************************
 * int node_swap_advertisements(context *ctx, int id, getquery *gq)
 * 	read the ads from the guy, and put our own in
 */

int node_swap_advertisements(context *ctx, int id, getquery *gq, double trust){
	int i;
	int list[MAXADVERTISE];
	int len;
	node *n;
	n = ctx->nodes[id];
	// discover new hosts from the query
	len = n->maxTargets;
	getquery_get_advertised(gq,list,&len);
	for(i=0;i<len;i++){// as long as we don't know they are bad
		if(trust_calc_back_cookie_path(ctx,n->id,list[i])>=0){
			// as them to our potentials list
			history_add_with_trust(ctx,n->id,list[i],trust);			
		}
	}
	// put our own hosts in the query
	getquery_advertised_reset(gq);
	getquery_advertise(gq,id);	// always advertise yourself
	for(i=0;i<n->nTargets;i++){
		if(n->targets[i]->trust > EPSILON)
			getquery_advertise(gq,n->targets[i]->id);
	}
	return 0;
}


/*********************************************************************
 * void node_handleget(struct context * ctx, struct event * e, double t);
 * 	handle a get request
 */

void node_handleget(struct context * ctx, struct event * e, double clock){
	getquery *gq;
	int max;
	int originalDesired;
	node * n;
	history *t;
	int willCheat=0;
	double coinflip;
	double pathTrust;
	int good;
	int len;

	gq = (getquery *) e->arg;
	fprintf(ctx->logfile,"%.30f : node_handleget:: got req from node %d to node %d for %d blocks\n",
			e->t, e->src_id,e->id, gq->init_want);
	n = ctx->nodes[e->id];
	assert(n->nMyBlocks<=n->nFreeSpace);

	pathTrust = trust_calc_cookie_path(ctx,e->id,gq);
	if(pathTrust<GOODTHRESH){	// not enough cookies
		good=trust_calc_back_cookie_path(ctx,e->src_id,e->id);// do the lookup ourselves
		if(good<0){	
			gq->init_want=gq->resp_want=0;	// tell them to bugger off
			event_create(ctx,node_handlegetresp,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
			e->arg=NULL;	// so we don't free the gq we just reused
			event_free(e);
			return;
		} 
		// else tell them to come back later with some probability
		coinflip = (double)random()/RAND_MAX;
		if(coinflip< ctx->backoffProb){
			gq->init_want=gq->resp_want=-1;	// tell them to come back later
			fprintf(ctx->logfile,"%.30f : node_handleget:: node %d told node %d to back off: path=%f\n",
					e->t, e->id,e->src_id, pathTrust);
			event_create(ctx,node_handlegetresp,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
			e->arg=NULL;	// so we don't free the gq we just reused
			event_free(e);
			return;
		}
		// else they are cool
	} else {
		// else, they are trusted, and we should pay attention to who the advertise
		node_swap_advertisements(ctx, e->id, gq, pathTrust);
	}
	t = history_lookup(ctx,e->id, e->src_id);
	//if(t && n->outstandingGets && t->shouldQuery==0){	// we have an outstanding query to them... 
	if(t && t->shouldQuery==0){	// we have an outstanding query to them... 
		gq->init_want=gq->resp_want=-1;	// tell them to come back later
		event_create(ctx,node_handlegetresp,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
		e->arg=NULL;	// so we don't free the gq we just reused
		event_free(e);
		return;
	}
	if(t)
		t->shouldQuery=0;	// remember not to query them again until checks
	// process the query
	len = MIN(n->nMyBlocks,gq->init_want);			// what we need to store
	assert(len>=0);
	max = history_get_max_transactions(ctx,t);		// if h=NULL, this returns 1
	// how much more we are willing to store
	max-=node_count_stored_blocks(ctx,e->id,e->src_id);	
	max=MAX(0,max);  	// don't allow negative values
	len=MIN(len,max);
	// fail/cheat with probability n->good
	coinflip=(double)random()/RAND_MAX;
	if((n->sleeperCount==0)&&(coinflip>n->good)) {
		willCheat=1;
	}
		
	if(n->sleeperCount>0)	// bide our time until they trust us with the precious...
		n->sleeperCount--;
	if(!ctx->cookieProtocol && t && t->badguy)	
		len=0;		// do no business with BAD folks if in strawman model
	originalDesired=gq->init_want;
	gq->resp_want=len;
	gq->init_want=len;
	// they are storing stuff for us
	if(gq->resp_want>0){
		// assert we have state space for them?
		assert(history_set_resp(ctx,e->id,e->src_id,gq->resp_want)!=-1);	
	}
	// and remember we are storing stuff for them
	if(gq->init_want>0){
		if(!willCheat)
			node_store_blocks(ctx,e->id,e->src_id,gq->init_want);
		else{
			if(ctx->nodes[e->src_id]->good>GOODTHRESH){	// only care if good ppl's jobs are dropped
				ctx->jobsDropped+=gq->init_want;
				ctx->nodes[e->src_id]->jobsDropped+=gq->init_want;
			}
				
		}
	}
	if(ctx->nodes[e->src_id]->good>GOODTHRESH){
		ctx->totJobs+=gq->init_want;
		ctx->nodes[e->src_id]->totJobs+=gq->init_want;
	}
	if(ctx->charity&&(len==0)&&(n->nMyBlocks==0)&&(max>0)&&(n->nFreeSpace>0)){	// charity case
		gq->init_want = MIN(max,n->nFreeSpace);
		gq->init_want = MIN(originalDesired,gq->init_want);
		if(!willCheat)		// cheating the charity case!?  Diabolical!
			node_store_blocks(ctx,e->id,e->src_id,gq->init_want);
	}
	n->nMyBlocks-=gq->resp_want;		// ASSUME the crypto makes them accept
	if(!willCheat)
		n->nFreeSpace-=gq->init_want;
	else {
		fprintf(ctx->logfile,"%.30f - node_handleget:: node %d CHEATED node %d out of %d blocks\n",
				e->t,e->id,e->src_id, gq->init_want);
	}
	event_create(ctx,node_handlegetresp,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
	e->arg=NULL;	// so we don't free the gq we just reused
	event_free(e);
}

/**********************************************
 * void node_handlegetresp(struct context * ctx, struct event * e, double t){
 * 	this is the message that tells us that how much the other side
 * 	has resp_want
 */
void node_handlegetresp(struct context * ctx, struct event * e, double clock){
	node *n;
	getquery *gq;
	checkquery *cq;
	int willCheat=0;
	double coinflip;
	trustcookie *tc;
	history *t;


	n = ctx->nodes[e->id];
	gq=(getquery *)e->arg;
	assert(n->nMyBlocks<=n->nFreeSpace);
	fprintf(ctx->logfile,"%.30f : node_handlegetRESP:: got resp from node %d "
			"to node %d accepting %d blocks; %d wanted\n",
			e->t, e->src_id,e->id, gq->init_want, gq->resp_want);
	tc = truststore_lookup(n->goodcookies,e->src_id);
	if(tc && tc->val>GOODTHRESH)	// if we trust them. then add their ads
		node_swap_advertisements(ctx, e->id, gq, tc->val);

	if((gq->resp_want==0)&&(gq->init_want==0)){
		fprintf(ctx->logfile,"%.30f : node_handlegetRESP:: node %d declared node %d useless\n",
				ctx->now,e->id,e->src_id);
		history_remove(ctx,e->id,e->src_id);	// they are useless to us, remove them
		n->outstandingGets--;
		t = history_lookup(ctx,e->id,e->src_id);
		if(t) 
			t->shouldQuery=1;		// mark them as we should come back
		if(n->outstandingGets==0)
			event_create(ctx,node_schedule,-1,e->id,NULL,ctx->now+EPSILON);
		event_free(e);
		return;
	}
	if((gq->init_want<=-1)||(gq->resp_want<=-1)){	// they told us to come back later
		n->outstandingGets--;
		t = history_lookup(ctx,e->id,e->src_id);
		if(t) 
			t->shouldQuery=1;		// mark them as we should come back
		if(n->outstandingGets==0)
			event_create(ctx,node_schedule,-1,e->id,NULL,ctx->now+EPSILON);
		event_free(e);
		return;
	}

	assert(n->nMyBlocks>=0);
	assert(n->nFreeSpace>=0);
	// accept the new blocks, unless we have already stored them elsewhere
	if(n->nMyBlocks<gq->init_want){	//we double stored some stuff, back out of part of the transaction
		cq = checkquery_create(e->id,gq->resp_want-n->nMyBlocks);
		event_create(ctx,node_handlejustkidding,e->id,e->src_id,cq,ctx->now+linkdelay(ctx));
		n->outstandingGets--;
		t = history_lookup(ctx,e->id,e->src_id);
		if(t) 
			t->shouldQuery=1;
		if(n->outstandingGets==0)
			event_create(ctx,node_schedule,-1,e->id,NULL,ctx->now+EPSILON);
		event_free(e);
		return;
	}
	n->nMyBlocks-=gq->init_want;
	assert(n->nMyBlocks>=0);
	// fail/cheat with probability n->good
	coinflip=(double)random()/RAND_MAX;
	if((n->sleeperCount==0)&&(coinflip>n->good)) {
		if(gq->init_want>0)	// only care if they've offered us something
			fprintf(ctx->logfile,"%.30f : node_handlegetresp:: node %d CHEATED node %d of %d blocks\n",
					e->t, e->id, e->src_id, gq->resp_want);
		willCheat=1;
		if(ctx->nodes[e->src_id]->good>GOODTHRESH){	// only care if good ppl's jobs are dropped
			ctx->jobsDropped+=gq->resp_want;
			ctx->nodes[e->src_id]->jobsDropped+=gq->resp_want;
		}
	}
	if(ctx->nodes[e->src_id]->good>GOODTHRESH){	// only care if good ppl's jobs are dropped
		ctx->totJobs+=gq->resp_want;
		ctx->nodes[e->src_id]->totJobs+=gq->resp_want;
	}
	if(n->sleeperCount>0)	// bide our time until they trust us with the precious...
		n->sleeperCount--;

	// remember how much we are storing for them
	if((gq->resp_want>0)&&(!willCheat)){	// we store for them as well
		node_store_blocks(ctx,e->id,e->src_id,gq->resp_want);
		n->nFreeSpace-=gq->resp_want;
	}
	// remember how much they are storing for us
	if((n->nMyBlocks>0)&&(gq->init_want>0)){		// still need to remember ifwe are cheating
		// in theory this should never ret -1, as we needed a history
		// entry for them to query for in the first place
		assert(history_set_resp(ctx,e->id,e->src_id,gq->init_want)!=-1);
	}
	n->outstandingGets--;		// dec reference count
	// fire off some queries to make sure they are REALLY holding our stuff
	if(gq->init_want>0){
		// one test now
		cq = checkquery_create(e->id,node_count_stored_blocks(ctx,e->id,e->src_id));
		event_create(ctx,node_handlecheck,e->id,e->src_id,cq,ctx->now+2*linkdelay(ctx));
	}
	assert(n->nMyBlocks>=0);
	assert(n->nFreeSpace>=0);
	if(n->outstandingGets != 0){
		event_free(e);
		return;		// wait until all of the responses are in
	}
	// fire off a reschedule if not satisfied
	if(n->nMyBlocks>0){
		event_create(ctx,node_schedule,-1,e->id,NULL,ctx->now+EPSILON);
	}else{
		node_print_status(ctx,n->id);
		n->done=1;
		if(n->good>GOODTHRESH)
			ctx->nSatisfied++;
	}
	
	event_free(e);
}

/***************************************************************
 *  void node_handlejustkidding(struct context * ctx, struct event * e, double t);
 *  	used when we over store stuff, and need to back out of a transaction
 *  		** note ** we are reusing the cq structure... so what
 */

void node_handlejustkidding(context * ctx, struct event * e, double t){
	node *n;
	checkquery *cq;
	int count;

	n = ctx->nodes[e->id];
	cq=(checkquery *)e->arg;
	fprintf(ctx->logfile,"%.30f : node_handlejustkidding:: got jk from node %d "
			"to %d re-accepting %d blocks\n",
			e->t, e->src_id,e->id, cq->nBlocks);
	if(ctx->nodes[e->src_id]->good>GOODTHRESH){
		ctx->totJobs-=cq->nBlocks;	// these jobs have beeb counted as good, so uncount 'em
		ctx->nodes[e->src_id]->totJobs+=cq->nBlocks;
	}
	count = node_count_stored_blocks(ctx,e->id,e->src_id);
	node_drop_blocks(ctx, e->id, e->src_id);
	count-=cq->nBlocks;
	if(count>0)
		node_store_blocks(ctx, e->id, e->src_id,count);
	n->nFreeSpace+=cq->nBlocks;		// fixed bug? 10/11 11:32am
	event_free(e);
}


/*****************************************************
 * void node_handlecheck(struct context * ctx, struct event * e, double t);
 * 	we have just recieved a request to make sure we aren't cheating
 */

void node_handlecheck(struct context * ctx, struct event * e, double t){
	node *n;
	checkquery *cq;
	int count;
		
	n = ctx->nodes[e->id];
	cq = (checkquery *)e->arg;
	assert(cq);
	fprintf(ctx->logfile,"%.30f : node_handlecheck:: req from node %d to node %d : us %d %d them %d %d \n",
			e->t,e->src_id, e->id,
			node_count_stored_blocks(ctx,e->id,e->src_id),
			history_get_resp(ctx,e->id,e->src_id),
			node_count_stored_blocks(ctx,e->src_id,e->id),
			history_get_resp(ctx,e->src_id,e->id)
			);
	cq->nBlocks = node_count_stored_blocks(ctx,e->src_id,e->id);	// HACK!
	count = history_get_resp(ctx,e->id,e->src_id);
	if(count <= cq->nBlocks){	// did they hold our stuff? // they might have gotten more
		history_update(ctx,e->id,e->src_id,1);
		// how much of their stuff to tell them we held
		// ASSUME crypto such that we can't lie
		cq->nBlocks = node_count_stored_blocks(ctx,e->id,e->src_id);
		cookie_give(ctx,e->id,e->src_id,1);
		event_create(ctx,node_handlecheckresp,e->id,e->src_id,cq,ctx->now+linkdelay(ctx));
		e->arg=NULL;	// set to null so the cq doesn't get free'd
		n->failAt=ctx->now+ctx->failTime;	// give 'em another 10 seconds
	} else {
		history_update(ctx,e->id,e->src_id,0);
		fprintf(ctx->logfile,"%.30f : node_handlecheck:: node %d says node %d is BAD : %d instead of %d\n",
				e->t,e->id, e->src_id, cq->nBlocks, count);
		// assert(ctx->nodes[e->src_id]->good<1);  // not a valid assertion - race condition
		// drop their shit and snub them
		if(ctx->nodes[e->id]->good>GOODTHRESH){
			ctx->jobsDropped+=cq->nBlocks;
			ctx->nodes[e->id]->jobsDropped+=cq->nBlocks;
		}
		n->nMyBlocks+=count;		// we need to replace the blocks we thought they had
		count=node_count_stored_blocks(ctx,e->id,e->src_id);
		n->nFreeSpace+=count;
		node_drop_blocks(ctx,e->id,e->src_id);	// drop their stuff
		history_set_abs_resp(ctx, e->id, e->src_id, 0);	// don't assume they have our stuff
		cookie_store_bad(ctx,e->id,e->src_id,0);
		if(ctx->cookieProtocol)
			history_remove(ctx,e->id,e->src_id);	// remove the offenders, they are in our badcookie list
		event_create(ctx,node_handlesnub,e->id,e->src_id,NULL,ctx->now+linkdelay(ctx));
	}
	event_free(e);
	return;
}


/*******************************************************
 * void node_handlecheckresp(struct context * ctx, struct event * e, double t);
 * 	we got a response to our check query... 
 */

void node_handlecheckresp(struct context * ctx, struct event * e, double t){
	node *n;
	checkquery *cq;
	int count;

	n = ctx->nodes[e->id];
	cq = (checkquery *)e->arg;
	assert(cq);
	fprintf(ctx->logfile,"%.30f : node_handlecheckresp:: resp from node %d to node %d : : us %d %d them %d %d \n",
					e->t,e->src_id, e->id,
			node_count_stored_blocks(ctx,e->id,e->src_id),
			history_get_resp(ctx,e->id,e->src_id),
			node_count_stored_blocks(ctx,e->src_id,e->id),
			history_get_resp(ctx,e->src_id,e->id)
			);
	cq->nBlocks = node_count_stored_blocks(ctx,e->src_id,e->id);	// HACK!
	count = history_get_resp(ctx,e->id,e->src_id);
	if(count <= cq->nBlocks){	// did they hold our stuff?	// -lt b/c they might have gotten more
		cookie_give(ctx,e->id,e->src_id,1);
		history_update(ctx,e->id,e->src_id,1);	// yes
		n->failAt=ctx->now+ctx->failTime;	// give 'em another 10 seconds
	} else {
		cookie_store_bad(ctx,e->id,e->src_id,0);	// store bad cookie ourselves
		history_update(ctx,e->id,e->src_id,0);	// no
		history_set_abs_resp(ctx, e->id, e->src_id, 0);	// don't assume they have our stuff
		if(ctx->cookieProtocol)
			history_remove(ctx,e->id,e->src_id);	// remove the offenders, they are in our badcookie list
		fprintf(ctx->logfile,"%.30f : node_handlecheckresp:: node %d says node %d is BAD : %d instead of %d\n",
				e->t,e->id, e->src_id, cq->nBlocks, count);
		// drop their shit and snub them
		//assert(ctx->nodes[e->src_id]->good<1);	// sometimes it is valid
		//						// if we have dropped all of their stuff
		//						// in the middle of processing a new request from them
		if(ctx->nodes[e->id]->good>GOODTHRESH){
			ctx->jobsDropped+=cq->nBlocks;
			ctx->nodes[e->id]->jobsDropped+=cq->nBlocks;
		}
		n->nMyBlocks+=count;		// we need to replace the blocks we thought they had
		count=node_count_stored_blocks(ctx,e->id,e->src_id);
		n->nFreeSpace+=count;
		node_drop_blocks(ctx,e->id,e->src_id);	// drop their stuff
		event_create(ctx,node_handlesnub,e->id,e->src_id,NULL,ctx->now+linkdelay(ctx));
	}
	event_free(e);
}

/************************************************************
 * void node_handlesnub(struct context * ctx, struct event * e, double t);
 * 	basically, we failed a file check, and they are telling us
 * 	to f'off, they are going to drop our stuff, and we should reciprocate
 */

void node_handlesnub(context * ctx, event * e, double t){
	node * n;
	int count;

	n = ctx->nodes[e->id];
	fprintf(ctx->logfile,"%.30f : node_handlesnub:: node %d got SNUBBED by node %d \n",
			e->t,e->id, e->src_id);
	history_update(ctx,e->id,e->src_id,0);	// buncha snubbing fuckers
	history_remove(ctx,e->id,e->src_id);	// buncha snubbing fuckers
	count = history_get_resp(ctx,e->id,e->src_id);
	n->nMyBlocks+=count;
	if(ctx->nodes[e->id]->good>GOODTHRESH){
		ctx->jobsDropped+=count;
		ctx->nodes[e->id]->jobsDropped+=count;
	}
	history_set_abs_resp(ctx, e->id, e->src_id, 0);	// don't assume they have our stuff
	count=node_count_stored_blocks(ctx,e->id,e->src_id);
	n->nFreeSpace+=count;
	node_drop_blocks(ctx,e->id,e->src_id);	// drop their stuff
	event_free(e);
}
