#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <string.h>

#include "truststore.h"
#include "link.h"

/* int truststore_init();
 * 	initialize anything that the truststore needs
 * 		- does not free resources from an existing truststore
 */

truststore * truststore_create(){
	truststore * ts;
	ts = malloc(sizeof(truststore));
	assert(ts);
	ts->root=NULL;
	ts->nodes=0;
	return ts;
}

/*************************************************************
 * trustcookie * trustcookie_create(int,int,double,int,int);
 * 	return an init'd trust cookie
 */

trustcookie * trustcookie_create(int src_id,int dst_id,double trust,int nTrans,int nGood){
	trustcookie *tc;

	tc= malloc(sizeof(trustcookie));
	assert(tc);
	tc->src_id=src_id;
	tc->dst_id=dst_id;
	tc->val=trust;
	tc->nTransactions=nTrans;
	tc->nGoodTransactions=nGood;
	return tc;
}


/* int truststore_add(trustcookie *);
 * 	adds a cookie to the store; return an error if it already exists
 */

int truststore_add(truststore *ts,trustcookie * cookie ){
	trustcookie * curr, * prev;
	int cmp;

	curr = ts->root;
	prev = NULL;
	ts->nodes++;

	cookie->left = cookie->right = NULL;
	cmp=0;

	while(curr){
		prev = curr;
		cmp = trustcookie_cmp(cookie, curr);
		if(cmp == 0){
			ts->nodes--;	// un-inc the counter
			return -1;	// exists already!!
		}
		if(cmp<0)
			curr= (trustcookie *) curr->left;
		else 
			curr= (trustcookie *) curr->right;
	}

	if(prev){
		if(cmp<0)	// we are the left child
			prev->left = (struct trustcookie *) cookie;
		else		// right child
			prev->right = (struct trustcookie *) cookie; 
	} else {
		ts->root = cookie;	// first node
	}

	return 0;
}

/*************************************************************************
 * 	truststore_remove(truststore *ts,trustcookie * cookie )
 * 		remove the cookie from the truststore
 */

int truststore_remove(truststore *ts, int id){
	trustcookie * curr, *prev;
	trustcookie * tcurr, *tprev;
	int cmp;
	int dir=0;

	prev=NULL;
	curr= ts->root;
	while(curr){
		cmp = id-curr->src_id;
		if(cmp == 0)
			break;
		if(cmp<0){
			dir=-1;	// left
			prev=curr;
			curr= curr->left;
		}else {
			dir=1;	// right;
			prev=curr;
			curr=curr->right;
		}
	}
	if(curr==NULL)	// didn't find it
		return -1;
	ts->nodes--;
	if(prev){	// if we are not deleting the root
		if(curr->left==NULL){	// the target has no left child
			if(dir==-1)
				prev->left=curr->right;
			else
				prev->right=curr->right;
		} else if(curr->right==NULL){ // the target has no right child
			if(dir==-1)
				prev->left=curr->left;
			else
				prev->right=curr->left;
		} else {	// complicated 2 child case
			tprev=NULL;
			tcurr=curr->right;
			while(tcurr->left!=NULL){
				tprev=tcurr;
				tcurr=tcurr->left;
			}	
			if(tprev){
				// delete replacement
				tprev->left=tcurr->right;
				tcurr->left=curr->left;
				tcurr->right=curr->right;
			} else { // tprev==NULL 
				tcurr->left=curr->left;
				// the right ptr stays put
			}
			if(dir==-1){	// replace the target node with the replacement
				prev->left=tcurr;
			} else {
				prev->right=tcurr;
			}

		}

	}else { // we are deleting the root
		if(curr->left==NULL){	// the target has no left child
			ts->root=curr->right;
		} else if(curr->right==NULL){ // the target has no right child
			ts->root=curr->left;
		} else {	// complicated 2 child case
			tprev=NULL;
			tcurr=curr->right;
			while(tcurr->left!=NULL){
				tprev=tcurr;
				tcurr=tcurr->left;
			}	
			if(tprev){
				// delete replacement
				tprev->left=tcurr->right;
				tcurr->left=curr->left;
				tcurr->right=curr->right;
			} else { // tprev==NULL 
				tcurr->left=curr->left;
				// the right ptr stays put
			}
			ts->root=tcurr;
		}
	}
	return 0;
}



/* int truststore_exists(trustcookie *);
 * 	return 1 if this cookie exists, 0 otherwise
 */

int truststore_exists(truststore *ts, int id){
	trustcookie * curr, * prev;
	int cmp;

	curr = ts->root;
	prev = NULL;
	while(curr){
		prev = curr;
		cmp = id-curr->src_id;
		if(cmp == 0)	return 1;	// found it
		if(cmp<0)
			curr= (trustcookie *) curr->left;
		else 
			curr= (trustcookie *) curr->right;
	}

	return 0;
}

/* trustcookie * truststore_lookup(nice_user * from);
 * 	looks up a cookie from the truststore based on who
 * 	it is from (in theory, all of the to: fields should
 * 	be to us)
 * 		uses the fact that cookies are first sorted by
 * 		the from field
 */

trustcookie * truststore_lookup(truststore *ts,int id ){
	trustcookie * curr;
	int cmp;

	curr = ts->root;
	while(curr){
		cmp = id - curr->src_id;
		if(cmp == 0)	return curr;	// found it
		if(cmp<0)
			curr= (trustcookie *) curr->left;
		else 
			curr= (trustcookie *) curr->right;
	}

	return NULL;
}
/**********************************************************
 * int trust_update_best_path(trustcookie *,int *,struct getquery *);
 * 	given paths a and b, over write a if b is better
 */

int trust_update_best_path(trustcookie * path,int *len,getquery * gq){
	double ma, mb;
	int i;

	ma = 9;
	mb = 9;
	assert(gq->queryid!=-1);
	for(i=0;i<*len;i++)
		ma = MIN(ma,path[i].val);
	for(i=0;i<gq->cookiePathLen;i++)
		mb = MIN(mb,gq->cookiePath[i].val);
	if(mb==9)
		return 0;	// path b is not valid; NOOP
	if(ma==9){		// path a is not valid, just copy
		*len=gq->cookiePathLen;
		for(i=0;i<*len;i++)
			path[i]=gq->cookiePath[i];
		return 0;
	}
	if(ma>mb)
		return 0;	//path a is better
	if(mb>ma){
		*len=gq->cookiePathLen;
		for(i=0;i<*len;i++)
			path[i]=gq->cookiePath[i];
		return 0;
	}
	// ma=mb
	if(*len<=gq->cookiePathLen)
		return 0;
	*len=gq->cookiePathLen;
	for(i=0;i<*len;i++)
		path[i]=gq->cookiePath[i];
	return 0;
}

/********************************************************************
 * int truststore_count(truststore*);
 * 	count the number of entries in the trust store
 * 	and make sure it matches the internal count
 */

int truststore_count1(trustcookie *tc){
	if(tc==NULL)
		return 0;
	return 1+truststore_count1(tc->left)+
		truststore_count1(tc->right);
}


int truststore_count(truststore*ts){
	int count;
	count=truststore_count1(ts->root);
	assert(count==ts->nodes);
	return 0;
}



/*********************************************************************/

int trustcookie_cmp(trustcookie *a, trustcookie *b){
	assert(a);
	assert(b);
	return a->src_id - b->src_id;
}
/********************************************************************
 * int cookie_give_good(struct context *, int src, int dst);
 * 	look up the person's cookie, add one if room, and increase
 * 	the good transaction count, and recalc the trust
 */

int cookie_give(context * ctx, int src_id, int dst_id,int good){
	node *dst;
	trustcookie * tc;

	if(!ctx->cookieProtocol)
		return 0;
	dst = ctx->nodes[dst_id];
	tc = truststore_lookup(dst->goodcookies,src_id);
	if(tc){	// good cookie already exist
		tc->nTransactions++;
		tc->nGoodTransactions+=good;
		assert(tc->nGoodTransactions!=0);
		tc->val=(double)tc->nGoodTransactions/tc->nTransactions;
		if(tc->val<GOODCOOKIETHRESH){	// cookie got worse
			truststore_remove(dst->goodcookies,src_id);
			truststore_add(dst->badcookies,tc);
			fprintf(ctx->logfile,"%f : cookie_give from %d to %d ; storing at %d : replacing GOOD with BAD cookie\n",
					ctx->now,src_id,dst_id, dst_id);
		} else
			fprintf(ctx->logfile,"%f : cookie_give from %d to %d ; storing at %d : updating GOOD cookie\n",
					ctx->now,src_id,dst_id, dst_id);

		return 0;
	}
	tc = truststore_lookup(dst->badcookies,src_id);
	if(tc){	// bad cookie already exists - but they did something right - weirdness
		tc->nTransactions++;
		tc->nGoodTransactions+=good;
		assert(tc->nGoodTransactions!=0);
		tc->val=(double)tc->nGoodTransactions/tc->nTransactions;
		if(tc->val>GOODCOOKIETHRESH){	// cookie got better
			truststore_remove(dst->badcookies,src_id);
			truststore_add(dst->goodcookies,tc);
			fprintf(ctx->logfile,"%f : cookie_give from %d to %d ; storing at %d : replacing BAD with GOOD cookie\n",
					ctx->now,src_id,dst_id, dst_id);
		} else
			fprintf(ctx->logfile,"%f : cookie_give from %d to %d ; storing at %d : updating BAD cookie\n",
					ctx->now,src_id,dst_id, dst_id);
		return 0;
	}
	// need to create a new cookie
	if(dst->maxCookies<=dst->goodcookies->nodes)
		return -1;	// not enough space to add cookie
	tc = trustcookie_create(src_id,dst_id,(double)good/1,good,1);
	fprintf(ctx->logfile,"%f : cookie_give from %d to %d ; storing at %d : new %s cookie\n",
			ctx->now,src_id,dst_id, dst_id,good?"GOOD":"BAD" );
	if(good)
		truststore_add(dst->goodcookies,tc);
	else 
		truststore_add(dst->badcookies,tc);
	return 0;
}

/********************************************************************
 * int cookie_store_bad(struct context *, int src, int dst);
 * 	save a bad cookie locally for the evil doer
 */

int cookie_store_bad(context * ctx, int src_id, int dst_id,int good){
	node *src;
	trustcookie * tc;

	if(!ctx->cookieProtocol)
		return 0;
	src = ctx->nodes[src_id];
	tc = truststore_lookup(src->badcookies,dst_id);
	if(tc){	// good cookie already exist
		tc->nTransactions++;
		tc->nGoodTransactions+=good;
		assert(tc->nGoodTransactions!=0);
		tc->val=(double)tc->nGoodTransactions/tc->nTransactions;
		return 0;
	}
	// need to create a new cookie
	if(src->maxCookies<=src->badcookies->nodes)
		return -1;	// not enough space to add cookie
	tc = trustcookie_create(dst_id,src_id,(double)good/1,good,1);	// store it backwards, so we can lookit up by dst
	fprintf(ctx->logfile,"%f : cookie_store_bad from %d to %d ; storing at %d\n",
			ctx->now,src_id,dst_id, src_id);
	truststore_add(src->badcookies,tc);
	return 0;
}

/***********************************************************************************
 * int trust_calc_cookie_path(struct context *,int,struct getquery *gq);
 * 	return 1 if the cookie path is good, else 0
 * 	USE a MIN rule for right now
 */

double trust_calc_cookie_path(context * ctx ,int src_id,getquery *gq){
	int i;
	double min=99;
	if(!ctx->cookieProtocol){		// with cookies off, always accept
		return 1.0;
	}
	for(i=0;i<gq->cookiePathLen;i++)
		min = MIN(min,gq->cookiePath[i].val);
	if(min==99.0)
		return EPSILON;
	else 
		return min;
}

/*************************************************************************************
 * int truststore_get_list(truststore *, trustcookie **, int *);
 * 	copy the BST into a list for ease of iteration
 */

int truststore_get_list1(trustcookie *tc, trustcookie **tc_list, int * index, int len){
	if(tc==NULL)
		return 0;
	if((*index)>=len)
		return 0;
	tc_list[(*index)++]=tc;
	return 1  
		+truststore_get_list1(tc->left,tc_list,index,len)
		+truststore_get_list1(tc->right,tc_list,index,len);

}

int truststore_get_list(truststore *ts, trustcookie **tc_list, int *len){
	int i;
	i=0;
	*len=truststore_get_list1(ts->root,tc_list,&i,*len);
	return 0;
}


/************************************************************************************
 * int trust_calc_forw_cookie_path(struct context *, int src, int dst);
 * 	start a cookie search from us (src) to them (dst)
 * 	- return 0 if we can locally retrieve the cookie, else return 1
 * 	- fill the info into the getquery
 * 	- mark them as unreachable in the mean time
 */

int trust_calc_forw_cookie_path(context *ctx, int src, int dst, getquery *gq){
	node *n;
	trustcookie *tc;
	history *h;

	if(!ctx->cookieProtocol){
		gq->cookiePathLen=0;
		return 0;
	}
	n = ctx->nodes[src];
	gq->src_id=src;
	gq->dst_id=dst;
	tc = truststore_lookup(n->goodcookies,dst);
	if(tc){	// we have a direct cookie from them
		gq->cookiePath[gq->cookiePathLen++]=*tc;
		return 0;
	}
	tc = truststore_lookup(n->badcookies,dst);
	if(tc){	// we have a direct bad cookie from them
		gq->cookiePath[gq->cookiePathLen++]=*tc;	// well... see if they like us anyway? 
		return -1;
	}

	h = history_lookup(ctx,src,dst);
	if(h->savedGetRequest){	// we've already calculated a cookie path
		gq->cookiePathLen=((getquery *)h->savedGetRequest)->cookiePathLen;
		// copy over saved cookiepath
		memcpy(gq->cookiePath,((getquery *)h->savedGetRequest)->cookiePath,
				sizeof(trustcookie)*gq->cookiePathLen);
		*gq=*(getquery*)h->savedGetRequest;
		free(h->savedGetRequest);
		h->savedGetRequest=NULL;
		return 0;
	}
	// need to actually query people for dst's cookies
	h->shouldQuery=1;
	// send ourselves a query lookup request  to save code
	gq->queryid=-1;
	gq->hopcount=ctx->MaxQueryHopCount;
	event_create(ctx,trust_cookielookup,n->id,n->id,gq,ctx->now+EPSILON);
	return 1;
}

/****************************************************************************
 * void trust_cookielookup(struct context *, event *, double t);
 * 	someone is trying to lookup a cookie through us
 */

void trust_cookielookup(context * ctx, event *e, double t){
	node * n;
	getquery * gq, *tmp_gq;
	truststore *ts;
	trustcookie **tc_list;
	trustcookie *tc;
	int i;
	int lqid;
	history *h;
	int len;
	n = ctx->nodes[e->id];
	
	tc_list= malloc(sizeof(trustcookie *)*n->maxCookies);
	assert(tc_list);
	gq=(getquery *)e->arg;
	gq->hopcount--;
	assert(ctx->cookieProtocol);

	//fprintf(ctx->logfile,"%f : trust_cookielookup:: at node %d from %d : %d -> %d : %d \n",
	//		ctx->now,e->id,e->src_id,gq->src_id,gq->dst_id, gq->hopcount);
	/*
	if(!gq->backLookup)
		ts=n->goodcookies;
	else
		ts=n->badcookies;
		*/
		ts=n->goodcookies;
	if((tc=truststore_lookup(ts,gq->dst_id))){
		gq->cookiePath[gq->cookiePathLen++]=*tc;
		event_create(ctx,trust_cookiereturn,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
		e->arg=NULL;
		event_free(e);
		free(tc_list);
		return;		// and we are done
	}
	if((n->nTransitQueries>=MAXTRANSITQUERIES)	// we have too many outstanding queries
			||(gq->hopcount<=0)){		// or ran out of hops
		gq->cookiePathLen=0;	// durh... we can't do anything
		event_create(ctx,trust_cookiereturn,e->id,e->src_id,gq,ctx->now+linkdelay(ctx));
		e->arg=NULL;
		event_free(e);
		free(tc_list);
		return;		// and we are done
	}
	n->transitQueries[n->nTransitQueries].id=n->qid++;		// mark this query id as used
	n->transitQueries[n->nTransitQueries].saved=gq->queryid;	// save the prev hop query id
	n->transitQueries[n->nTransitQueries].prevhop=e->src_id;	// save the prev hop query id
	n->transitQueries[n->nTransitQueries].queriesSent=0;		// reset q count
	n->transitQueries[n->nTransitQueries].bestLen=0;		// reset q count
	if(gq->queryid==-1)	// if we originated this query, fixup the reverse path
		n->transitQueries[n->nTransitQueries].saved=n->transitQueries[n->nTransitQueries].id;
	gq->queryid=n->transitQueries[n->nTransitQueries].id;
	lqid=n->nTransitQueries;
	n->nTransitQueries++;
	// try to route to the digests
	getquery_mark_visited(gq,n->id);
	len = n->maxCookies;
	truststore_get_list(n->goodcookies,tc_list,&len);
	for(i=0;i<len;i++){
		if(getquery_already_visited(gq,tc_list[i]->dst_id))
			continue;
		//assume b/c we have their cookies, we can check their digests
		// if they know the dst directly, send it to them
		tc = truststore_lookup(ctx->nodes[tc_list[i]->dst_id]->goodcookies,gq->dst_id);
		if(tc!=NULL){
			tmp_gq = getquery_copy(gq);
			tmp_gq->cookiePath[tmp_gq->cookiePathLen++]=*tc_list[i];	// tack the next hop on
			event_create(ctx,trust_cookielookup, e->id,tc_list[i]->dst_id, tmp_gq, ctx->now+linkdelay(ctx));
			n->transitQueries[lqid].queriesSent++;
		}
		if(n->transitQueries[lqid].queriesSent>=ctx->CookieSearchExpansionFactor){
			event_free(e);
			free(tc_list);
			return;	// we have done all we can
		}
			
	}
	// if we have gotten here, we don't have any cookies for them and don't know of
	// any who has cookies, just query some ppl at random
	// well... the history list is already randomized, so use that
	for(i=0;i<len;i++){
		if(getquery_already_visited(gq,tc_list[i]->dst_id))
			continue;
		tmp_gq = getquery_copy(gq);
		tmp_gq->cookiePath[tmp_gq->cookiePathLen++]=*tc_list[i];	// tack the next hop on
		event_create(ctx,trust_cookielookup, e->id,tc_list[i]->dst_id, tmp_gq, ctx->now+linkdelay(ctx));
		n->transitQueries[lqid].queriesSent++;
		if(n->transitQueries[lqid].queriesSent>=ctx->CookieSearchExpansionFactor){
			event_free(e);
			free(tc_list);
			return;	// we have done all we can
		}
	}

	if(n->transitQueries[lqid].queriesSent==0){	// didn't send any queries
		if(gq->src_id != e->id){ // tell them the bad news
			tmp_gq=getquery_copy(gq);
			tmp_gq->cookiePathLen=0;
			tmp_gq->queryid=n->transitQueries[lqid].saved;
			event_create(ctx,trust_cookiereturn,e->id,e->src_id,tmp_gq,ctx->now+linkdelay(ctx));
		} else {	// it's just a failed local lookup; don't bother sending messages, just doit
			h = history_lookup(ctx,gq->src_id,gq->dst_id);
			if(h){	// we could have removed them in the time being, so if they aren't here, just ignore
				// the cookie return thinger
				h->savedGetRequest=gq;	// cache our query results so we find them next time
				h->shouldQuery=1;
				e->arg=NULL;
			}
			
		}
	}
	free(tc_list);
	event_free(e);
}

/************************************************************************************************
 * void trust_cookiereturn(struct context *, event *, double t);
 * 	process the returned cookie paths
 * 	- only pass stuff back up to the prev hop if this was our last query
 * 	- only return the best one to the next upstream path
 * 	- best as defined by MAX MIN, and then shortest for a tie breaker
 * 	- if we are the origin of the query, do some additional processing
 */

void trust_cookiereturn(context * ctx, event *e, double t){
	node *n;
	getquery *gq;
	int lqid;
	int i;
	int prev;
	history *h;

	assert(ctx->cookieProtocol);
	n = ctx->nodes[e->id];
	gq=(getquery *) e->arg;
	e->arg=NULL;

	//fprintf(ctx->logfile,"%f : trust_cookiereturn:: at node %d from %d : %d -> %d : %d \n",
	//		ctx->now,e->id,e->src_id,gq->src_id,gq->dst_id, gq->hopcount);
	if(gq->queryid==-1){	// NULL querry has returned, just shortcut back
		h = history_lookup(ctx,gq->src_id,gq->dst_id);
		if(h){	// we could have removed them in the time being, so if they aren't here, just ignore
			// the cookie return thinger
			h->savedGetRequest=gq;	// cache our query results so we find them next time
			h->shouldQuery=1;
			e->arg=NULL;
		}
		event_free(e);
		return;
	}

	for(i=0;i<n->nTransitQueries;i++)
		if(n->transitQueries[i].id == gq->queryid)
			break;
	assert(i<n->nTransitQueries);
	lqid=i;

	trust_update_best_path(n->transitQueries[lqid].best,
			&n->transitQueries[lqid].bestLen,
			gq);
	n->transitQueries[lqid].queriesSent--;
	if(n->transitQueries[lqid].queriesSent>0){	// just wait for more returns
		free(gq);
		event_free(e);	
		return;	
	}
	// copy over best query
	gq->queryid=n->transitQueries[lqid].saved;
	gq->cookiePathLen=n->transitQueries[lqid].bestLen;
	prev = n->transitQueries[lqid].prevhop;		// get prev hop
	for(i=0;i<gq->cookiePathLen;i++)	// copy over best path
		gq->cookiePath[i]=n->transitQueries[lqid].best[i];

	// clean up the transit query info stuff
	n->nTransitQueries--;
	n->transitQueries[lqid]=n->transitQueries[n->nTransitQueries];
	
	if(n->id != gq->src_id){	// if we didn't originate this query
		event_create(ctx,trust_cookiereturn,e->id,prev,gq,ctx->now+linkdelay(ctx));
		e->arg=NULL;
		event_free(e);
		return;
	}
	// else we are the dst for the returned cookie package
	h = history_lookup(ctx,gq->src_id,gq->dst_id);
	if(h){	// we could have removed them in the time being, so if they aren't here, just ignore
		// the cookie return thinger
		h->savedGetRequest=gq;	// cache our query results so we find them next time
		h->shouldQuery=1;
		e->arg=NULL;
	}
	event_free(e);
}

/*********************************************************************************
 * int trust_calc_back_cookie_path(struct context *, int src, int dst);
 * 	<0  ---> known bad people
 * 	>0  ---> known good people
 * 	0   ---> have to do remote search, come back later; NOT IMPLEMENTED
 * 		just assume good unless proven otherwise
 */
int trust_calc_back_cookie_path(context * ctx, int src, int dst){
	//getquery *tmp;
	trustcookie *tc;
	node *n;
	int i;
	trustcookie ** tc_list;
	int len;

	if(!ctx->cookieProtocol)
		return 1;

	n = ctx->nodes[src];
	tc = truststore_lookup(n->badcookies, dst);
	if(tc) 
		return -1;
	len = n->maxCookies;
	tc_list= malloc(sizeof(trustcookie *)*n->maxCookies);
	assert(tc_list);
	truststore_get_list(n->goodcookies,tc_list,&len);
	for(i=0;i<len;i++){		// asume b/c we have a cookie from them
		if(truststore_lookup(ctx->nodes[tc_list[i]->dst_id]->badcookies,dst))
			return -1;	// my friend thinks they are bad
	}
	free(tc_list);
	return 1;	// no info, defaulting to GOOD
}



void truststore_print(trustcookie *tc,int depth, FILE *f){
	int i;
	if(tc==NULL)
		return;
	for(i=0;i<depth;i++)
		fprintf(f,"\t");
	fprintf(f,"%d\n",tc->src_id);
	truststore_print(tc->left,depth+1,f);
	truststore_print(tc->right,depth+1,f);
	return;
}


int do_truststore_tests(){
	truststore * ts;
	trustcookie * tc[20];
	int i;
	ts = truststore_create();
	assert(ts);


	for(i=0;i<20;i++){
		tc[i] = trustcookie_create(i,-1,1.0,1,1);
	}

	truststore_add(ts,tc[2]);
	truststore_add(ts,tc[1]);
	truststore_add(ts,tc[3]);
	truststore_add(ts,tc[4]);
	truststore_count(ts);
	truststore_remove(ts,2);
	truststore_count(ts);
	truststore_remove(ts,4);
	truststore_count(ts);
	truststore_remove(ts,3);
	truststore_count(ts);
	truststore_remove(ts,1);
	assert(ts->nodes==0);
	truststore_count(ts);


	truststore_add(ts,tc[10]);
	truststore_add(ts,tc[5]);
	truststore_add(ts,tc[15]);
	truststore_add(ts,tc[3]);
	truststore_add(ts,tc[7]);
	truststore_add(ts,tc[13]);
	truststore_add(ts,tc[17]);
	truststore_add(ts,tc[2]);
	truststore_add(ts,tc[4]);
	truststore_add(ts,tc[12]);
	truststore_add(ts,tc[14]);

	truststore_remove(ts,15);
	truststore_count(ts);
	truststore_print(ts->root,0,stdout);
	truststore_remove(ts,10);
	truststore_print(ts->root,0,stdout);
	truststore_count(ts);
	truststore_remove(ts,17);
	truststore_count(ts);
	truststore_remove(ts,12);
	truststore_count(ts);




	ts = truststore_create();
	for(i=0;i<20;i++){
		truststore_add(ts,tc[i]);
	}
	assert(ts->nodes==20);
	assert(truststore_remove(ts,3)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,14)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,2)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,12)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,4)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,11)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,19)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,8)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,9)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,0)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,10)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,1)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,13)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,7)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,15)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,16)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,17)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,5)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,18)==0);
	truststore_count(ts);
	assert(truststore_remove(ts,6)==0);
	truststore_count(ts);
	assert(ts->nodes==0);


	return 0;
}
